package eu.ztsh.garmin

import android.annotation.SuppressLint
import android.bluetooth.BluetoothAdapter
import android.bluetooth.BluetoothDevice
import android.bluetooth.BluetoothSocket
import android.util.Log
import com.mapbox.navigation.tripdata.maneuver.model.Maneuver
import eu.ztsh.garmin.data.DataCache
import eu.ztsh.garmin.data.GarminManeuver
import eu.ztsh.garmin.data.GarminMapper
import eu.ztsh.garmin.data.GarminModelItem
import eu.ztsh.garmin.data.MapboxMapper
import java.io.IOException
import java.util.*
import java.util.concurrent.SynchronousQueue

@SuppressLint("MissingPermission")
class Garmin(
    val context: MainActivity,
    val device: BluetoothDevice,
    val adapter: BluetoothAdapter
) {

    private lateinit var connection: ConnectThread
    private val observer = ProcessingThreadObserver()
    private val cache = DataCache()

    fun start() {
        connection = ConnectThread()
        connection.start()
    }

    fun close() {
        connection.close()
    }

    fun process(maneuver: Maneuver) {
        ManeuverProcessingThread(maneuver).start()
    }

//    fun process(location: LocationMatcherResult) {
//        LocationProcessingThread(location).start()
//    }

//    fun process(navigationSessionState: NavigationSessionState) {
//        cache.update(navigationSessionState)
//    }

    private inner class ManeuverProcessingThread(val maneuver: Maneuver) : ProcessingThread<GarminManeuver>() {

        override fun process(): GarminManeuver? {
            if (cache.hasChanged(maneuver)) {
                cache.update(maneuver)
                return MapboxMapper.map(maneuver)
            }
            return null
        }

        override fun enqueue(item: GarminManeuver) {
            if (cache.hasChanged(item.lanes)) {
                connection.enqueue(GarminMapper.map(item.lanes))
            }
            if (cache.hasChanged(item.direction)) {
                connection.enqueue(GarminMapper.map(item.direction))
            }
            if (cache.hasChanged(item.distance)) {
                connection.enqueue(GarminMapper.map(item.distance))
            }
            // flag?
        }

    }

//    private inner class LocationProcessingThread(val location: LocationMatcherResult) : ProcessingThread() {
//
//        override fun process() {
//            if (cache.hasChanged(location)) {
//                cache.update(location)
//                send(MapboxMapper.apply(location))
//            }
//        }
//    }

    private abstract inner class ProcessingThread<T : GarminModelItem> : Thread() {

        abstract fun process(): T?

        abstract fun enqueue(item: T)

        override fun run() {
            val processing = process()
            if (processing != null) {
                enqueue(processing)
                cache.update(processing)
            }
            observer.join(this)
        }

    }

    private inner class ProcessingThreadObserver {

        fun <T : GarminModelItem> join(thread: ProcessingThread<T>) {
            thread.join()
        }

    }

    private inner class ConnectThread : Thread() {

        private val queue: SynchronousQueue<IntArray> = SynchronousQueue()
        private var current: IntArray = intArrayOf()

        private val socket: BluetoothSocket? by lazy(LazyThreadSafetyMode.NONE) {
            context.checkBt()
            device.createRfcommSocketToServiceRecord(UUID.fromString("00001101-0000-1000-8000-00805F9B34FB"))
        }

        override fun run() {
            // Cancel discovery because it otherwise slows down the connection.
            context.checkBt()
            adapter.cancelDiscovery()
            try {
                socket?.connect()
                context.setConnectionStatus(true)
                sleep(3000)
                readAll()
                while (true) {
                    val newCurrent = queue.poll()
                    if (newCurrent == null) {
                        sleep(500)
                    } else {
                        current = newCurrent
                    }
                    send(current)
                }
            } catch (e: IOException) {
                Log.d(TAG, "Not connected", e)
                context.setConnectionStatus(false)
                while (true) {
                    // Just dequeue
                    // TODO: Add option to reconnect
                    queue.poll()
                    sleep(900)
                }
            }
        }

        // Closes the client socket and causes the thread to finish.
        fun close() {
            try {
                socket?.close()
            } catch (e: IOException) {
                Log.e(TAG, "Could not close the client socket", e)
            }
        }

        fun enqueue(data: IntArray) {
            queue.put(data)
        }

        private fun send(data: IntArray) {
            sendRaw(prepareData(data))
        }

        private fun readAll() {
            val buffer = ByteArray(64)
            val istr = socket!!.inputStream
            istr!!.let {
                if (it.available() > 0) {
                    it.read(buffer)
                }
            }
        }

        private fun sendRaw(buff: IntArray) {
            buff.forEach { socket!!.outputStream.write(it) }
            socket!!.outputStream.flush()
            sleep(2000)
            readAll()
        }

    }

    companion object {

        lateinit var instance: Garmin

        fun prepareData(input: IntArray): IntArray {
            val n = input.size
            var crc = (0xeb + n + n).toUInt()
            val chars = ArrayList<Int>()
            chars.add(0x10)
            chars.add(0x7b)
            chars.add((n + 0x06))
            if (n == 0xa)
                chars.add(0x10)
            chars.add(n)
            chars.add(0x00)
            chars.add(0x00)
            chars.add(0x00)
            chars.add(0x55)
            chars.add(0x15)
            for (char in input) {
                crc = (crc + char.toUInt())
                chars.add(char)
                if (char == 0x10)
                    chars.add(0x10)
            }
            chars.add((-(crc.toInt()) and 0xff))
            chars.add(0x10)
            chars.add(0x03)
            return chars.toIntArray()
        }

        private const val TAG = "GARMIN"
    }
}
